==> dvote-0.7.0.tar.gz: lib/net/gateway.dart <==
import 'dart:convert';
import 'dart:async';
import 'dart:math';
import 'dart:typed_data';
import 'package:dvote/util/dev.dart';
import 'package:flutter/foundation.dart';
import 'package:http/http.dart';
import '../util/json-sign.dart';
import "../constants.dart";

import 'package:web_socket_channel/io.dart';
import 'package:web3dart/web3dart.dart';

var _random = Random.secure();

// ----------------------------------------------------------------------------
// Exported classes
// ----------------------------------------------------------------------------

/// Client class to send WS requests to a DVote Gateway
class DVoteGateway {
  String _uri;
  String _publicKey;

  /// The WebSocket "open" channel
  IOWebSocketChannel _channel;
  List<DVoteGatewayRequest> _requests = [];

  // /// List of callbacks to invoke when an unrelated message is received.
  // ObserverList<Function> _listeners = new ObserverList<Function>();

  DVoteGateway(this._uri, {String publicKey}) {
    this._publicKey = publicKey;
    connect(this._uri);
  }

  bool get isConnected => _channel != null;
  String get uri => _uri;
  String get publicKey => _publicKey;

  /// Initialization the WebSockets connection with the server
  connect(String newUri, {bool disconnectOld = false}) {
    if (newUri == null || newUri == "") throw Exception("Invalid Gateway URI");

    if (isConnected && disconnectOld) disconnect();

    _channel = new IOWebSocketChannel.connect(newUri);
    _channel.stream.listen(_gotMessage);
  }

  /// Disconnects the WebSocket communication
  disconnect() {
    if (_channel != null) {
      if (_channel.sink != null) {
        _channel.sink.close();
      }
      _channel = null;
    }
  }

  /// Perform a raw request to the Vocdoni Gateway and wait for a response to
  /// arrive within a given timeout
  Future<Map<String, dynamic>> request(Map<String, dynamic> body,
      {int timeout = 30, String privateKey}) async {
    if (_channel == null || _channel.sink == null)
      throw Exception("You are not connected to a gateway");

    final id = makeRandomId();

    Completer<Map<String, dynamic>> comp = Completer<Map<String, dynamic>>();
    DVoteGatewayRequest req = DVoteGatewayRequest(id, comp,
        originalRequest: kReleaseMode ? "" : jsonEncode(body));
    _requests.add(req);

    Future.delayed(Duration(seconds: timeout)).then((_) {
      if (comp.isCompleted) return;
      comp.completeError(Exception("The request timed out"));
    });

    Map<String, dynamic> requestBody = Map.from(body);
    if (!(requestBody["timestamp"] is int)) {
      requestBody["timestamp"] =
          (DateTime.now().millisecondsSinceEpoch / 1000).floor();
    }

    Map<String, dynamic> content;

    // Sign if needed
    if (privateKey == null || privateKey == "") {
      content = {"id": id, "request": requestBody, "signature": ""};
    } else {
      final String signature = await signJsonPayload(requestBody, privateKey);
      content = {"id": id, "request": requestBody, "signature": signature};
    }

    _channel.sink.add(jsonEncode(content));

    return comp.future;
  }

  /// Try to match the incoming message to a previous request
  _gotMessage(message) async {
    Uint8List data = Uint8List.fromList(message);
    var incomingMessage;

    try {
      incomingMessage = jsonDecode(String.fromCharCodes(data));
    } catch (err) {
      return devPrint("ERR: Received a non-JSON message");
    }

    if (incomingMessage == null) {
      return devPrint("ERR: Received an empty message");
    } else if (!(incomingMessage is Map)) {
      return devPrint("ERR: Received an invalid message");
    } else if (incomingMessage["id"] == null) {
      return _handleUnrelatedMessage(incomingMessage);
    }

    final req = _requests.firstWhere((req) => req.id == incomingMessage["id"],
        orElse: () => null);
    if (req == null) {
      // NO ORIGINARY REQUEST
      return devPrint("ERR: Received a message for a non-existing request: " +
          incomingMessage["id"]);
    }

    _requests.remove(req);

    if (req.completer.isCompleted) {
      // ALREADY HANDLED
      return devPrint(
          "Got a response for an already completed (timed out) request: ${req.originalRequest}");
    }

    final from = (new DateTime.now().millisecondsSinceEpoch / 1000).floor() -
        SIGNATURE_TIMESTAMP_TOLERANCE;
    final until = (new DateTime.now().millisecondsSinceEpoch / 1000).floor() +
        SIGNATURE_TIMESTAMP_TOLERANCE;

    if (incomingMessage["error"] != null) {
      Map error;
      final err = incomingMessage["error"];
      if (err is String)
        error = jsonDecode(err);
      else
        error = incomingMessage["error"];

      // ERROR RESPONSE
      if (!(error is Map)) {
        return req.completer
            .completeError(Exception("Received a non-object 'error' response"));
      } else if (incomingMessage["id"] != error["request"]) {
        return req.completer.completeError(
            Exception("The signed request ID does not match the expected one"));
      }

      if (!(error["timestamp"] is int) ||
          error["timestamp"] < from ||
          error["timestamp"] > until) {
        return req.completer
            .completeError(Exception("The response timestamp is invalid"));
      } else if (!(await isValidJsonSignature(
          incomingMessage["signature"], error, publicKey))) {
        return req.completer
            .completeError(Exception("The response signature is not valid"));
      } else if (!(error["message"] is String)) {
        return req.completer
            .completeError(Exception("Received an empty error message"));
      }
      return req.completer.completeError(Exception(error["message"]));
    }

    if (incomingMessage["response"] == null) {
      return req.completer
          .completeError(Exception("Received an empty response"));
    }
    final response = incomingMessage["response"];

    if (!(response is Map)) {
      return req.completer
          .completeError(Exception("Received an invalid response"));
    }

    // CHECK RESPONSE
    if (incomingMessage["id"] != response["request"]) {
      return req.completer.completeError(
          Exception("The signed request ID does not match the expected one"));
    }
    if (!(response["timestamp"] is int) ||
        response["timestamp"] < from ||
        response["timestamp"] > until) {
      return req.completer
          .completeError(Exception("The response timestamp is invalid"));
    } else if (!(await isValidJsonSignature(
        incomingMessage["signature"], response, publicKey))) {
      return req.completer
          .completeError(Exception("The response signature is not valid"));
    }

    // DONE
    if (req.completer.isCompleted) {
      return devPrint(
          "Received a response for an already completed (timed out) request");
    }
    req.completer.complete(response);
  }

  // /// Adds a callback to be invoked in case of incoming notification
  // addListener(Function callback) {
  //   _listeners.add(callback);
  // }

  // removeListener(Function callback) {
  //   _listeners.remove(callback);
  // }

  _handleUnrelatedMessage(Map incomingMessage) {
    devPrint("GOT AN UNRELATED MESSAGE: " + incomingMessage.toString());
    // _listeners.forEach((Function callback) {
    //   callback(message);
    // });
  }
}

/// Client class to wrap calls to Ethereum Smart Contracts using a (Vocdoni) Gateway
class Web3Gateway {
  String rpcUri;
  String wsUri;
  Web3Client client;

  Web3Gateway(String gatewayUri) {
    connect(gatewayUri);
  }

  /// Initialization the WebSockets connection with the Gateway
  connect(String gatewayUri) {
    if (gatewayUri == null || gatewayUri == "")
      throw Exception("Invalid Gateway URI");
    else if (gatewayUri.startsWith("http")) {
      this.rpcUri = gatewayUri;
      this.wsUri = gatewayUri.replaceFirst(RegExp("^http"), "ws");
    } else if (gatewayUri.startsWith("ws")) {
      this.rpcUri = gatewayUri.replaceFirst(RegExp("^ws"), "http");
      this.wsUri = gatewayUri;
    }

    disconnect();
    client = Web3Client(rpcUri, Client(), socketConnector: () {
      return IOWebSocketChannel.connect(wsUri).cast<String>();
    });
  }

  /// Disconnects the WebSocket communication
  disconnect() {
    if (client != null) {
      client.dispose();
      client = null;
    }
  }

  /// Perform a call request to Ethereum
  Future<List<dynamic>> callMethod(String contractAbi, String contractAddress,
      String method, List<dynamic> params) async {
    if (client == null) throw Exception("You are not connected to Ethereum");

    final ethAddress = EthereumAddress.fromHex(contractAddress);
    final contract = DeployedContract(
        ContractAbi.fromJson(contractAbi, 'Vocdoni'), ethAddress);

    final methodFunc = contract.function(method);
    try {
      final result = await client.call(
          contract: contract, function: methodFunc, params: params ?? []);
      return result;
    } on FormatException catch (err) {
      devPrint("ERROR: $err");
      throw Exception("Invalid response from the Web3 Gateway: $rpcUri");
    } catch (err) {
      throw err;
    }
  }

  /// Send a transaction to the blockchain
  Future<String> sendTransaction(String contractAbi, String contractAddress,
      String method, List<dynamic> params, Credentials credentials) {
    if (client == null)
      throw Exception("You are not connected to Ethereum");
    else if (credentials == null)
      throw Exception("Credentials are required to send a transaction");

    final ethAddress = EthereumAddress.fromHex(contractAddress);
    final contract = DeployedContract(
        ContractAbi.fromJson(contractAbi, 'Vocdoni'), ethAddress);

    final methodFunc = contract.function(method);

    return client.sendTransaction(
        credentials,
        Transaction.callContract(
            contract: contract,
            function: methodFunc,
            parameters: params ?? []));
  }
}

// ----------------------------------------------------------------------------
// Internal classes
// ----------------------------------------------------------------------------

class DVoteGatewayRequest {
  String id;
  Completer completer;
  String originalRequest; // debug only

  DVoteGatewayRequest(this.id, this.completer, {this.originalRequest});
}

String makeRandomId() {
  final values = List<int>.generate(16, (i) => _random.nextInt(256));
  return base64.encode(values);
}

==> dvote-0.7.1.tar.gz: lib/net/gateway.dart <==
import 'dart:convert';
import 'dart:async';
import 'dart:math';
import 'dart:typed_data';
import 'package:dvote/util/dev.dart';
import 'package:flutter/foundation.dart';
import 'package:http/http.dart';
import 'package:web_socket_channel/web_socket_channel.dart';
import '../util/json-sign.dart';
import "../constants.dart";

import 'package:web_socket_channel/io.dart';
import 'package:web3dart/web3dart.dart';

final _random = Random.secure();

/// Invoke `_onTimeout` when `TIMEOUT_COUNT_THRESHOLD` requests expire
/// within a period of `TIMEOUT_TIME_FRAME` seconds
const TIMEOUT_COUNT_THRESHOLD = 2;
const TIMEOUT_TIME_FRAME = Duration(seconds: 15);

// ----------------------------------------------------------------------------
// Exported classes
// ----------------------------------------------------------------------------

/// Client class to send WS requests to a DVote Gateway
class DVoteGateway {
  String _uri;
  String _publicKey;

  /// The WebSocket "open" channel
  IOWebSocketChannel _channel;
  List<DVoteGatewayRequest> _requests = [];

  /// Callback to invoke when N attempts expire for a period of T seconds
  void Function() _onTimeout;
  List<DateTime> _timeoutTimestamps = [];
  Timer _timeoutCheckTimer;

  // /// List of callbacks to invoke when an unrelated message is received.
  // ObserverList<Function> _listeners = new ObserverList<Function>();

  DVoteGateway(this._uri, {String publicKey, void Function() onTimeout}) {
    this._publicKey = publicKey;
    this._onTimeout = onTimeout ?? () {};
    connect(this._uri);
  }

  bool get isConnected => _channel != null;
  String get uri => _uri;
  String get publicKey => _publicKey;

  /// Initialization the WebSockets connection with the server
  connect(String newUri, {bool disconnectOld = false}) {
    if (newUri == null || newUri == "") throw Exception("Invalid Gateway URI");

    if (isConnected && disconnectOld) disconnect();

    _channel = new IOWebSocketChannel.connect(newUri,
        pingInterval: Duration(seconds: 1));
    _channel.stream.listen(_gotMessage);

    // Check the timeout rate every second
    _timeoutCheckTimer =
        Timer.periodic(Duration(seconds: 1), (_) => _checkTimeout());
  }

  /// Disconnects the WebSocket communication
  disconnect() {
    if (_channel != null) {
      if (_channel.sink != null) {
        _channel.sink.close();
      }
      _channel = null;
    }

    if (_timeoutCheckTimer is Timer) {
      if (_timeoutCheckTimer.isActive) _timeoutCheckTimer.cancel();
    }
  }

  /// Closes the current connection and attempts to connect again
  reconnect(Object err) {
    if (err != null) devPrint("GW Error: $err");
    disconnect();
    connect(_uri, disconnectOld: true);
  }

  /// Perform a raw request to the Vocdoni Gateway and wait for a response to
  /// arrive within a given timeout
  Future<Map<String, dynamic>> request(Map<String, dynamic> body,
      {int timeout = 30, String privateKey}) async {
    if (_channel == null || _channel.sink == null)
      throw Exception("You are not connected to a gateway");

    final id = makeRandomId();

    Completer<Map<String, dynamic>> comp = Completer<Map<String, dynamic>>();
    DVoteGatewayRequest req = DVoteGatewayRequest(id, comp,
        originalRequest: kReleaseMode ? "" : jsonEncode(body));
    _requests.add(req);

    // Trigger a timeout after N seconds
    Future.delayed(Duration(seconds: timeout)).then((_) {
      if (comp.isCompleted) return;
      comp.completeError(Exception("The request timed out"));
      _reportTimeout(req.created);
    });

    Map<String, dynamic> requestBody = Map.from(body);
    if (!(requestBody["timestamp"] is int)) {
      requestBody["timestamp"] =
          (DateTime.now().millisecondsSinceEpoch / 1000).floor();
    }

    Map<String, dynamic> content;

    // Sign if needed
    if (privateKey == null || privateKey == "") {
      content = {"id": id, "request": requestBody, "signature": ""};
    } else {
      final String signature = await signJsonPayload(requestBody, privateKey);
      content = {"id": id, "request": requestBody, "signature": signature};
    }

    _channel.sink.add(jsonEncode(content));

    return comp.future;
  }

  /// Try to match the incoming message to a previous request
  _gotMessage(message) async {
    // Handle potential error events
    if (message is WebSocketChannelException) {
      devPrint("GW: Disconnected. Reconnecting.");
      reconnect(message);
      return;
    } else if (!(message is List<int>)) {
      return devPrint(
          "GW: Received an event of type ${message.runtimeType}. Skipping");
    }

    Uint8List data = Uint8List.fromList(message);
    var incomingMessage;

    try {
      incomingMessage = jsonDecode(String.fromCharCodes(data));
    } catch (err) {
      return devPrint("ERR: Received a non-JSON message");
    }

    if (incomingMessage == null) {
      return devPrint("ERR: Received an empty message");
    } else if (!(incomingMessage is Map)) {
      return devPrint("ERR: Received an invalid message");
    } else if (incomingMessage["id"] == null) {
      return _handleUnrelatedMessage(incomingMessage);
    }

    final req = _requests.firstWhere((req) => req.id == incomingMessage["id"],
        orElse: () => null);
    if (req == null) {
      // NO ORIGINARY REQUEST
      return devPrint("ERR: Received a message for a non-existing request: " +
          incomingMessage["id"]);
    }

    _requests.remove(req);

    if (req.completer.isCompleted) {
      // ALREADY HANDLED
      return devPrint(
          "Got a response for an already completed (timed out) request: ${req.originalRequest}");
    }

    final from = (new DateTime.now().millisecondsSinceEpoch / 1000).floor() -
        SIGNATURE_TIMESTAMP_TOLERANCE;
    final until = (new DateTime.now().millisecondsSinceEpoch / 1000).floor() +
        SIGNATURE_TIMESTAMP_TOLERANCE;

    if (incomingMessage["error"] != null) {
      Map error;
      final err = incomingMessage["error"];
      if (err is String)
        error = jsonDecode(err);
      else
        error = incomingMessage["error"];

      // ERROR RESPONSE
      if (!(error is Map)) {
        return req.completer
            .completeError(Exception("Received a non-object 'error' response"));
      } else if (incomingMessage["id"] != error["request"]) {
        return req.completer.completeError(
            Exception("The signed request ID does not match the expected one"));
      }

      if (!(error["timestamp"] is int) ||
          error["timestamp"] < from ||
          error["timestamp"] > until) {
        return req.completer
            .completeError(Exception("The response timestamp is invalid"));
      } else if (!(await isValidJsonSignature(
          incomingMessage["signature"], error, publicKey))) {
        return req.completer
            .completeError(Exception("The response signature is not valid"));
      } else if (!(error["message"] is String)) {
        return req.completer
            .completeError(Exception("Received an empty error message"));
      }
      return req.completer.completeError(Exception(error["message"]));
    }

    if (incomingMessage["response"] == null) {
      return req.completer
          .completeError(Exception("Received an empty response"));
    }
    final response = incomingMessage["response"];

    if (!(response is Map)) {
      return req.completer
          .completeError(Exception("Received an invalid response"));
    }

    // CHECK RESPONSE
    if (incomingMessage["id"] != response["request"]) {
      return req.completer.completeError(
          Exception("The signed request ID does not match the expected one"));
    }
    if (!(response["timestamp"] is int) ||
        response["timestamp"] < from ||
        response["timestamp"] > until) {
      return req.completer
          .completeError(Exception("The response timestamp is invalid"));
    } else if (!(await isValidJsonSignature(
        incomingMessage["signature"], response, publicKey))) {
      return req.completer
          .completeError(Exception("The response signature is not valid"));
    }

    // DONE
    if (req.completer.isCompleted) {
      return devPrint(
          "Received a response for an already completed (timed out) request");
    }
    req.completer.complete(response);
  }

  // /// Adds a callback to be invoked in case of incoming notification
  // addListener(Function callback) {
  //   _listeners.add(callback);
  // }

  // removeListener(Function callback) {
  //   _listeners.remove(callback);
  // }

  _handleUnrelatedMessage(Map incomingMessage) {
    devPrint("GOT AN UNRELATED MESSAGE: " + incomingMessage.toString());
    // _listeners.forEach((Function callback) {
    //   callback(message);
    // });
  }

  _reportTimeout(DateTime createdTime) {
    _timeoutTimestamps.add(createdTime);
  }

  _checkTimeout() {
    final now = DateTime.now();
    _timeoutTimestamps = _timeoutTimestamps
        .where((item) {
          final diff = item.difference(now).abs();
          return diff.inMilliseconds < TIMEOUT_TIME_FRAME.inMilliseconds;
        })
        .cast<DateTime>()
        .toList();

    if (_timeoutTimestamps.length >= TIMEOUT_COUNT_THRESHOLD) {
      _timeoutTimestamps = [];
      _onTimeout();

      // Already notified => stop checking
      if (_timeoutCheckTimer.isActive) _timeoutCheckTimer.cancel();
      _timeoutCheckTimer = null;
    }
  }
}

/// Client class to wrap calls to Ethereum Smart Contracts using a (Vocdoni) Gateway
class Web3Gateway {
  String rpcUri;
  String wsUri;
  Web3Client client;

  Web3Gateway(String gatewayUri) {
    connect(gatewayUri);
  }

  /// Initialization the WebSockets connection with the Gateway
  connect(String gatewayUri) {
    if (gatewayUri == null || gatewayUri == "")
      throw Exception("Invalid Gateway URI");
    else if (gatewayUri.startsWith("http")) {
      this.rpcUri = gatewayUri;
      this.wsUri = gatewayUri.replaceFirst(RegExp("^http"), "ws");
    } else if (gatewayUri.startsWith("ws")) {
      this.rpcUri = gatewayUri.replaceFirst(RegExp("^ws"), "http");
      this.wsUri = gatewayUri;
    }

    disconnect();
    client = Web3Client(rpcUri, Client(), socketConnector: () {
      return IOWebSocketChannel.connect(wsUri).cast<String>();
    });
  }

  /// Disconnects the WebSocket communication
  disconnect() {
    if (client != null) {
      client.dispose();
      client = null;
    }
  }

  /// Perform a call request to Ethereum
  Future<List<dynamic>> callMethod(String contractAbi, String contractAddress,
      String method, List<dynamic> params) async {
    if (client == null) throw Exception("You are not connected to Ethereum");

    final ethAddress = EthereumAddress.fromHex(contractAddress);
    final contract = DeployedContract(
        ContractAbi.fromJson(contractAbi, 'Vocdoni'), ethAddress);

    final methodFunc = contract.function(method);
    try {
      final result = await client.call(
          contract: contract, function: methodFunc, params: params ?? []);
      return result;
    } on FormatException catch (err) {
      devPrint("ERROR: $err");
      throw Exception("Invalid response from the Web3 Gateway: $rpcUri");
    } catch (err) {
      throw err;
    }
  }

  /// Send a transaction to the blockchain
  Future<String> sendTransaction(String contractAbi, String contractAddress,
      String method, List<dynamic> params, Credentials credentials) {
    if (client == null)
      throw Exception("You are not connected to Ethereum");
    else if (credentials == null)
      throw Exception("Credentials are required to send a transaction");

    final ethAddress = EthereumAddress.fromHex(contractAddress);
    final contract = DeployedContract(
        ContractAbi.fromJson(contractAbi, 'Vocdoni'), ethAddress);

    final methodFunc = contract.function(method);

    return client.sendTransaction(
        credentials,
        Transaction.callContract(
            contract: contract,
            function: methodFunc,
            parameters: params ?? []));
  }
}

// ----------------------------------------------------------------------------
// Internal classes
// ----------------------------------------------------------------------------

class DVoteGatewayRequest {
  DateTime created;
  String id;
  Completer completer;
  String originalRequest; // debug only

  DVoteGatewayRequest(this.id, this.completer, {this.originalRequest}) {
    created = DateTime.now();
  }
}

String makeRandomId() {
  final values = List<int>.generate(16, (i) => _random.nextInt(256));
  return base64.encode(values);
}
